use std::sync::Arc;
use std::time::Duration;

use crd::NeonDatabase;
use futures::stream::StreamExt;
use k8s_openapi::apiextensions_apiserver::pkg::apis::apiextensions::v1::CustomResourceDefinition;
use kube::api::Api;
use kube::api::Patch;
use kube::api::PatchParams;
use kube::runtime::conditions;
use kube::runtime::wait::await_condition;
use kube::runtime::watcher::Config;
use kube::Client;
use kube::CustomResourceExt;
use kube::Resource;
use kube::ResourceExt;
use kube::{runtime::controller::Action, runtime::Controller};
use tracing::error;
use tracing::info;
use tracing::instrument;

mod crd;
mod minio;
mod neon;

struct ContextData {
    pub kube_client: Client,
}

impl ContextData {
    pub fn new(kube_client: Client) -> Self {
        Self { kube_client }
    }
}

/// All errors possible to occur during reconciliation
#[derive(Debug, thiserror::Error)]
pub enum Error {
    /// Any error originating from the `kube-rs` crate
    #[error("Kubernetes reported error: {source}")]
    KubeError {
        #[from]
        source: kube::Error,
    },
    /// Error in user input or NeonDatabase resource definition, typically missing fields.
    #[error("Invalid NeonDatabase CRD: {0}")]
    UserInputError(String),
    #[error("General error: {0}")]
    GeneralError(String),
}

#[tokio::main]
async fn main() {
    tracing_subscriber::fmt::init();
    let kube_client = Client::try_default()
        .await
        .expect("Failed to create client");

    let crd_api: Api<NeonDatabase> = Api::all(kube_client.clone());
    let context: Arc<ContextData> = Arc::new(ContextData::new(kube_client.clone()));

    let ssapply = PatchParams::apply("neon_operator").force();
    let crds: Api<CustomResourceDefinition> = Api::all(kube_client.clone());
    info!(
        "Creating crd: {}",
        serde_json::to_string(&NeonDatabase::crd()).unwrap()
    );
    crds.patch(
        "neondatabases.melenion.com",
        &ssapply,
        &Patch::Apply(NeonDatabase::crd()),
    )
    .await
    .expect("Patching the CRD failed");

    info!("Waiting for the api-server to accept the CRD");
    let establish = await_condition(
        crds,
        "neondatabases.melenion.com",
        conditions::is_crd_established(),
    );
    let _ = tokio::time::timeout(std::time::Duration::from_secs(10), establish)
        .await
        .expect("Timeout waiting for CRD to establish");

    info!("CRD established, starting controller");
    Controller::new(crd_api.clone(), Config::default())
        .run(reconcile, on_error, context)
        .for_each(|rec_result| async move {
            match rec_result {
                Ok(o) => {
                    println!("Went OK {o:?}")
                }
                Err(e) => {
                    eprintln!("Reconciliation error:\n{:?}", e);
                }
            }
        })
        .await;
}

#[instrument(err, fields(database_name = ?neon.metadata.name), skip(neon, context))]
async fn reconcile(neon: Arc<NeonDatabase>, context: Arc<ContextData>) -> Result<Action, Error> {
    let client = context.kube_client.clone();

    let namespace: String = match neon.namespace() {
        Some(ns) => ns,
        None => return Err(Error::UserInputError("Missing namespace".to_string())),
    };

    let owner_ref = neon
        .controller_owner_ref(&())
        .ok_or(Error::GeneralError("Missing owner ref".to_string()))?;

    minio::create_deployment(client.clone(), &owner_ref, &namespace).await?;
    minio::create_service(client.clone(), &owner_ref, &namespace).await?;
    neon::reconcile_storage_broker(
        client.clone(),
        &owner_ref,
        &namespace,
        &neon.spec.neon_image_ref,
    )
    .await?;
    neon::reconcile_page_server(
        client.clone(),
        &owner_ref,
        &namespace,
        &neon.spec.neon_image_ref,
    )
    .await?;
    neon::reconcile_safe_keepers(
        client.clone(),
        &owner_ref,
        &namespace,
        &neon.spec.neon_image_ref,
    )
    .await?;
    neon::reconcile_compute(
        client.clone(),
        &owner_ref,
        &namespace,
        &neon.spec.compute_image_ref,
    )
    .await?;
    Ok(Action::requeue(Duration::from_secs(5)))
}

fn on_error(db: Arc<NeonDatabase>, error: &Error, _context: Arc<ContextData>) -> Action {
    error!("Reconciliation error:\n{:?}.\n{:?}", error, db);
    Action::requeue(Duration::from_secs(5))
}
